/* Reaction Time Tester
 * by mikalhart - mhcom <at> sundial <dot> com
 *
 * A game that measures a person's reaction time
 *
 * You press the big red button to start play. 
 * When the green light comes on, you release the 
 * button as quickly as possible, and the RTT displays 
 * the number of milliseconds that it took you to react. 
 * Repeat 10 times, and your "score" for the "game" is the 
 * sum of the 10 individual response times. 
 *
 * Your goal, of course, is to try to minimize your 
 * cumulative score. At the end of the game, the top 10 
 * all-time scores (stored in EEPROM) are displayed on the 
 * 7-segment display panel. If you are lucky enough to get 
 * the low score, LEDs flash admiringly. 
 *
 * http://www.arduino.cc/playground/ComponentLib/ReactionTimeTester
 */
 
#include <EEPROM.h>
#include <HC4LED.h>

// Input pin
#define BUTTON_PIN 8

// Output LEDs
#define CHEAT_LED 12
#define READY_LED 11
#define POWER_LED 13
#define GO_LED 10

// LCD display lines
#define DISP_DATA_PIN 4
#define DISP_CLOCK_PIN 5
#define DISP_BLANK_PIN 7

// Change this to customize the owner
#define OWNER "GABE"

// Change to "true" to enable serial debugging
#define DEBUG false

// This trick makes all the embedded serial stuff go away magically
#if !DEBUG
#define Serial NullSerial
class NullSerialClass
{
  public:
  void begin(int baud) {}
  void println(int x, int type=DEC) {}
  void println(const char *p=NULL) {}
  void print(int x, int type=DEC) {}
  void print(const char *p) {}
  int available() {return 0;}
  int read() {return -1;}
  void flush() {}
} NullSerial;
#endif

// Global variables
bool bCheat = false, bRecord = false;

// Set the LED state depending on period and duty
void flash_led(int pin, int period, int duty)
{
  int cycle = millis() % period;
  digitalWrite(pin, cycle <= duty ? HIGH : LOW);
}

// Wait until a pin reaches a certain logic level
void wait_for_pin(int pin, int LEVEL)
{
  while (digitalRead(pin) == !LEVEL);
}

// Returns true if pin attained LEVEL, or false if 
// time expired.  During the wait, the cheat, ready and go
// LEDs are flash appropriately
bool wait_for_pin_or_timer(int pin, int LEVEL, unsigned long ms)
{
  unsigned long start = millis();
  unsigned long cur;
  do
  {
    if (bCheat) flash_led(CHEAT_LED, 400, 200);
    if (bRecord) 
    {
      flash_led(READY_LED, 400, 200);
      flash_led(GO_LED, 400, 200);
    }
    cur = millis();
  }
  while (digitalRead(pin) == !LEVEL && (unsigned)(cur - start) < ms);

  bool ret = (unsigned)(cur - start) < ms;
  if (ret)
  {
    Serial.print("WFPT: "); Serial.print(pin); Serial.print(LEVEL); Serial.println(ms);
    Serial.print("Button pressed!  cur: ");
    Serial.print(cur);
    Serial.print(", start: ");
    Serial.println(start);
  }
  return ret;
}

// Flash a "hello!"
void flash_all_leds(int interval, int count)
{
  for (int i=0; i<count; ++i)
  {
    digitalWrite(POWER_LED, HIGH);
    digitalWrite(CHEAT_LED, HIGH);
    digitalWrite(READY_LED, HIGH);
    digitalWrite(GO_LED, HIGH);
    
    delay(interval);

    digitalWrite(POWER_LED, LOW);
    digitalWrite(CHEAT_LED, LOW);
    digitalWrite(READY_LED, LOW);
    digitalWrite(GO_LED, LOW);
    
    delay(interval);
  }
}

// The HC4LED interface object
HC4LED disp(DISP_CLOCK_PIN, DISP_DATA_PIN, DISP_BLANK_PIN, true);

void setup()
{
  pinMode(BUTTON_PIN, INPUT);
  pinMode(CHEAT_LED, OUTPUT);
  pinMode(READY_LED, OUTPUT);
  pinMode(POWER_LED, OUTPUT);
  pinMode(GO_LED, OUTPUT);

  digitalWrite(BUTTON_PIN, HIGH);
  
  disp.clear();

  Serial.begin(9600);
  Serial.println("Begin");
  
  // If user has his finger inserted at startup, then this is a special cue to erase the EEPROM
  if (digitalRead(BUTTON_PIN) == LOW)
  {
    for (int i=0; i<512; ++i)
      EEPROM.write(i, 0xFF);
    Serial.println("EEPROM erased!");
    flash_all_leds(100, 4);
    disp.scroll_text("ERASE   ", 1000);
    delay(2000);
    
    // require that he lift his finger before beginning (to get a good seed)
    while (digitalRead(BUTTON_PIN) == LOW);
  }
  
  // Hello, world!
  flash_all_leds(400, 2);
  
  disp.scroll_text(OWNER "'S TOY    ", 0);
}

void loop()
{
  unsigned long start, delta;
  unsigned short top_nine_scores[9];
  unsigned long total_score;
  
  // Load the top nine high (low!) scores for display
  for (int i=0; i<9; ++i)
    top_nine_scores[i] = (byte)EEPROM.read(2*i) + 256 * (byte)EEPROM.read(2*i + 1);
    
  // During the "inactive" state scroll through high scores until button is pushed
  bool lets_play = false;
  while (!lets_play)
  {
    for (int i=0; i<9; ++i)
    {
      if (top_nine_scores[i] == 0 || top_nine_scores[i] == 0xFFFF)
        break;
      disp.display_number(i + 1, 0);
      if (lets_play = wait_for_pin_or_timer(BUTTON_PIN, LOW, 1000))
        break;
      disp.display_number(top_nine_scores[i], 0); 
      if (lets_play = wait_for_pin_or_timer(BUTTON_PIN, LOW, 3000))
        break;
    }
    if (!lets_play)
      lets_play = wait_for_pin_or_timer(BUTTON_PIN, LOW, 100);
  }
  
  // Reset all LEDs
  flash_all_leds(2, 250);
  digitalWrite(POWER_LED, HIGH);
  
  // Begin game with "PLAY" message on display
  disp.scroll_text("PLAY", 2000);
  bCheat = bRecord = false;
  
  // Wait for user to lift "start" button
  wait_for_pin(BUTTON_PIN, HIGH);

  // Randomize
  randomSeed(millis());
  
  total_score = 0;

  // Repeat the test 10 times
  for (int count = 0; count < 10; ++count)
  {
    // Turn off go LED
    digitalWrite(GO_LED, LOW);

    disp.display_number(count + 1, 0);

    // Wait for button press and debounce
    while (digitalRead(BUTTON_PIN) == HIGH)
      flash_led(READY_LED, 100, 90);

    // debounce button for 200ms
    for (start = millis(); (unsigned long)(millis() - start) < 200;)
      flash_led(READY_LED, 100, 90);
    
    // Start a counter for between 2 and 5 seconds
    delta = random(2000, 5000);

    // Wait for the timer to expire
    for (start = millis(); (unsigned long)(millis() - start) < delta; )
    {
      // Flash the "record" lamp
      flash_led(READY_LED, 100, 90);
      
      // Did he release the button before the timer expired??
      if (digitalRead(BUTTON_PIN) == HIGH)
      {
        Serial.println("Cheat!");
        digitalWrite(CHEAT_LED, HIGH);
        digitalWrite(READY_LED, LOW);
        disp.scroll_text("CHEAT    ", 0);
        bCheat = true;
        return;
      }
    }
  
    // At the end of timer, light the "go" light
    Serial.println("GO!");
    digitalWrite(READY_LED, LOW);
    digitalWrite(GO_LED, HIGH);
  
    // Wait for player to remove finger: watch for "light touch" inverse debounce
    int score;
    bool bButtonLifted = false;
    while (!bButtonLifted)
    {
      wait_for_pin(BUTTON_PIN, HIGH);
      score = (unsigned long)(millis() - start - delta);
      bButtonLifted = true;
      
      // Make sure button is up at least 200ms
      for (long i=millis(); bButtonLifted && (unsigned long)(millis() - i) < 200; )
        if (digitalRead(BUTTON_PIN) == LOW)
          bButtonLifted = false;
    }
    
    digitalWrite(GO_LED, LOW);
    
    // Calculate score
    Serial.print("Score is "); Serial.println(score);
    total_score += score;
    disp.display_number(score > 9999 ? 9999 : score, 1000);
  }

  // display score
  if (total_score > 9999) total_score = 9999;
  Serial.print("Cumulative score is "); Serial.println(total_score);
  disp.display_number(total_score, 2000);
  
  // Insert new score into record list
  int cur = total_score;
  for (int i=0; i<256; ++i)
  {
    unsigned short record = (byte)EEPROM.read(2*i) + 256 * (byte)EEPROM.read(2*i + 1);
    
    if (record == 0 || cur < record)
    {
      // If it's a new record, then display a special message
      if (i == 0) // first place!
      {
        Serial.print("A new world record: "); Serial.print(cur); Serial.println("!");
        digitalWrite(READY_LED, HIGH);
        disp.scroll_text("BEST SCORE    ", 0);
        bRecord = true;
      }

      // otherwise, display the rank of the new score
      // but only if it's in the top 9
      if (cur == total_score && i < 9)
      {
        if (i == 0)
          disp.scroll_text("FIRST PLACE    ", 250);
        else if (i == 1)
          disp.scroll_text("SECOND PLACE    ", 250);
        else if (i == 2) 
          disp.scroll_text("THIRD PLACE    ", 250);
        else
        {
          char string[] = "cTH PLACE    ";
          string[0] = i + '1';
          disp.scroll_text(string, 250);
        }
      }
          
      EEPROM.write(2 * i, (byte)cur);
      EEPROM.write(2 * i + 1, (byte)(cur / 256));
      Serial.print(i + 1); Serial.print(". "); Serial.print(cur);
      Serial.println(cur == total_score ? " <---" : "");
      cur = record;
    }
    else
    {
      Serial.print(i + 1); Serial.print(". "); Serial.println(record);
    }
  
    if (record == 0 || record == 0xFFFF)
      break;
  }
}
